local addon, options, L, ext = unpack(select(2, ...))
local module = addon:CreateModule("commandBar")

local DEBUG = ext.DEBUG
local DEBUGF = ext.DEBUGF
local tfind = ext.tfind
local tmap = ext.tmap
local secondsToString = ext.secondsToString

local CURRENT_PLAYER = MissionMinder.CURRENT_PLAYER

---------------------------------------------
-- LOCALE
---------------------------------------------
local L_NOT_STARTED = L["Not started"];

---------------------------------------------
-- Command Bar
---------------------------------------------


local function createNoticeIcons(self)
	self:SetWidth(self:GetWidth() + 5 + 16) -- Adjust frame width to account for the new icons

	self.ReadyIcon = self:CreateTexture(nil, "ARTWORK")
	self.ReadyIcon:SetPoint("LEFT", self.Count, "RIGHT", 5, 0)
	self.ReadyIcon:SetSize(16, 16)
	self.ReadyIcon:SetTexture("Interface\\FriendsFrame\\Raf_Textures")
	self.ReadyIcon:SetTexCoord(0, 25/512, 0, 25/512)
	self.ReadyIcon:Hide()

	self.TrainingIcon = self:CreateTexture(nil, "ARTWORK")
	self.TrainingIcon:SetPoint("CENTER", self.ReadyIcon)
	self.TrainingIcon:SetSize(20, 20)
	self.TrainingIcon:SetTexture("Interface\\COMMON\\StreamCircle")
	self.TrainingIcon:SetVertexColor(1, 0.82, 0)
	self.TrainingIcon:Hide()

	self.NotStartedIcon = self:CreateTexture(nil, "ARTWORK")
	self.NotStartedIcon:SetPoint("CENTER", self.ReadyIcon)
	self.NotStartedIcon:SetSize(16, 16)
	self.NotStartedIcon:SetTexture("Interface\\ENCOUNTERJOURNAL\\UI-EJ-WarningTextIcon")
	self.NotStartedIcon:Hide()

	self.Anim = self:CreateAnimationGroup()
	self.Anim:SetLooping("REPEAT")
	self.Anim.rotate = self.Anim:CreateAnimation("Rotation")
	self.Anim.rotate:SetChildKey("TrainingIcon")
	self.Anim.rotate:SetOrder(1)
	self.Anim.rotate:SetDuration(1)
	self.Anim.rotate:SetDegrees(-360)
	self.Anim:Play()
end

local trainingFinishedTimer = nil
local function displayExtendedTroopDetails(self, ...)
	if ( trainingFinishedTimer ) then
		trainingFinishedTimer:Cancel()
		trainingFinishedTimer = nil
	end

	local troops = tmap(C_Garrison.GetFollowerShipments(LE_GARRISON_TYPE_7_0), function(key, containerID)
		local name, _, _, ready, total = C_Garrison.GetLandingPageShipmentInfoByContainerID(containerID);
		return name, {
			ready = ready;
			training = total - ready;
		}
	end)

	-- This code is largely based on OrderHallCommandBarMixin:RefreshCategories()
	self.categoryPool:ReleaseAll();
	local categoryInfo = C_Garrison.GetClassSpecCategoryInfo(LE_FOLLOWER_TYPE_GARRISON_7_0);

	local numCategories = #categoryInfo;
	local prevCategory, firstCategory;
	local xSpacing = 20;	-- space between categories
	for index, category in ipairs(categoryInfo) do
		local categoryInfoFrame = self.categoryPool:Acquire();
		categoryInfoFrame.Icon:SetTexture(category.icon);
		categoryInfoFrame.Icon:SetTexCoord(0, 1, 0.25, 0.75);
		categoryInfoFrame.name = category.name;
		categoryInfoFrame.description = category.description;

		categoryInfoFrame.UpdateTooltip = categoryInfoFrame:GetScript("OnEnter")

		if ( not categoryInfoFrame.ReadyIcon ) then
			createNoticeIcons(categoryInfoFrame)
		end

		local openSlots =  category.limit - category.count;
		local troop = troops[category.name] or {}

		local ready = troop.ready or 0
		local training = troop.training or 0
		local availableToStart = openSlots - ready - training

		categoryInfoFrame.Count:SetFormattedText(ORDER_HALL_COMMANDBAR_CATEGORY_COUNT, category.count, category.limit);
		categoryInfoFrame.ReadyIcon:Hide()
		categoryInfoFrame.TrainingIcon:Hide()
		categoryInfoFrame.NotStartedIcon:Hide()
		if ( openSlots <= 0 ) then
			categoryInfoFrame.Count:SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		elseif ( ready > 0 ) then
			categoryInfoFrame.Count:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			categoryInfoFrame.ReadyIcon:Show()
		elseif ( training > 0 ) then
			categoryInfoFrame.Count:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			categoryInfoFrame.TrainingIcon:Show()

			-- Start timer to force category update when we expect it to finish
			local now = time()
			local workorder = tfind(MissionMinderData.Characters[CURRENT_PLAYER].followerWorkOrders, category.name, "name")
			trainingFinishedTimer = C_Timer.NewTimer(max(workorder.endDate - now, 1), C_Garrison.RequestLandingPageShipmentInfo)
		elseif ( availableToStart > 0 ) then -- Have not started
			categoryInfoFrame.Count:SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
			categoryInfoFrame.NotStartedIcon:Show()
		end

		categoryInfoFrame:ClearAllPoints();
		if (not firstCategory) then
			-- calculate positioning so that the set of categories ends up being centered
			categoryInfoFrame:SetPoint("LEFT", self, "RIGHT", (0 - (numCategories * (categoryInfoFrame:GetWidth() + xSpacing))) - 30, 0);
			firstCategory = categoryInfoFrame;
		else
			categoryInfoFrame:SetPoint("LEFT", prevCategory, "RIGHT", xSpacing, 0);
		end
		categoryInfoFrame:Show();
		prevCategory = categoryInfoFrame;
	end
end


function fetchTroopDetails(troopName)
	local troops = {}
	local followers = C_Garrison.GetFollowers(LE_FOLLOWER_TYPE_GARRISON_7_0) or {}
	for i = 1, #followers do
		local follower = followers[i]
		if ( follower.isTroop and follower.name == troopName ) then
			table.insert(troops, follower)
		end
	end

	local summary = {}
	for i = 1, #troops do
		local details = {
			id = troops[i].followerID;
			name = troops[i].name;
			maxDurability = troops[i].maxDurability;
			currentDurability=  troops[i].durability;
			durabilityLoss = 0;
		}

		if ( troops[i].status == GARRISON_FOLLOWER_ON_MISSION ) then -- Calculate durability loss
			details.durabilityLoss = 1
		end
		table.insert(summary, details)
	end

	table.sort(summary,  function(summary1, summary2)
		if ( summary1.currentDurability ~= summary2.currentDurability ) then
			return summary1.currentDurability > summary2.currentDurability -- Higher currentDurability first
		elseif( summary1.durabilityLoss ~= summary2.durabilityLoss ) then
			return summary1.durabilityLoss < summary2.durabilityLoss -- Lower durabilityLoss first
		else
			return summary1.id < summary2.id
		end
	end)
	return summary;
end



MissionMinderCommandBarTroopDurabilityTooltipInsetMixIn = {}
function MissionMinderCommandBarTroopDurabilityTooltipInsetMixIn:SetDurability(durability, maxDurability, durabilityLoss)
	local heartWidth = 13;
	local spacing = -2;

	durability = Clamp(durability, 0, maxDurability);
	durabilityLoss = Clamp(durabilityLoss or 0, 0, durability);
	durability = durability - durabilityLoss;

	while ((self.durability and #self.durability or 0) < maxDurability) do
		local durabilityTexture = self:CreateTexture(nil, "ARTWORK", "GarrisonMissionFollowerButtonDurabilityTemplate");
		durabilityTexture:ClearAllPoints();
		if (#self.durability == 1) then
			durabilityTexture:SetPoint("TOPRIGHT");
		else
			durabilityTexture:SetPoint("TOPRIGHT", self.durability[#self.durability - 1], "TOPLEFT", spacing, 0);
		end
	end

	for i = 1, durability do
		self.durability[i]:Show();
		self.durability[i]:SetAtlas("GarrisonTroops-Health");
		self.durability[i]:SetDesaturated(false);
	end
	for i = durability + 1, durability + durabilityLoss do
		self.durability[i]:Show();
		self.durability[i]:SetAtlas("GarrisonTroops-Health-Consume");
		self.durability[i]:SetDesaturated(false);
	end
	for i = durability + durabilityLoss + 1, maxDurability do
		self.durability[i]:Show();
		self.durability[i]:SetAtlas("GarrisonTroops-Health");
		self.durability[i]:SetDesaturated(true);
	end
	for i = maxDurability + 1, (self.durability and #self.durability or 0) do
		self.durability[i]:Hide();
	end
end

local durabilityFrames = CreateFramePool("FRAME", self, "MissionMinderCommandBarTroopDurabilityTooltipInset")

local function showExtendedTroopDetails(self, ...)
	durabilityFrames:ReleaseAll()

	local categoryInfo = C_Garrison.GetClassSpecCategoryInfo(LE_FOLLOWER_TYPE_GARRISON_7_0);
	local category = tfind(categoryInfo, self.name, "name");


	local openSlots = category.limit - category.count
	local readyTroops, trainingTroops, availableToStart = 0, 0, openSlots
	if ( openSlots > 0) then
		local followerShipments = C_Garrison.GetFollowerShipments(LE_GARRISON_TYPE_7_0);
		for i = 1, #followerShipments do
			local name, _, _, ready, total = C_Garrison.GetLandingPageShipmentInfoByContainerID(followerShipments[i])
			if ( self.name == name ) then
				readyTroops = ready
				trainingTroops = total - ready
				availableToStart = openSlots - readyTroops - trainingTroops
				break;
			end
		end
	end

	local troopDetails = fetchTroopDetails(category.name)
	GameTooltip:AddLine(" ")
	for i = 1, #troopDetails do
		local frame = durabilityFrames:Acquire()
		frame.Name:SetText(troopDetails[i].name)
		frame:SetDurability(troopDetails[i].currentDurability, troopDetails[i].maxDurability, troopDetails[i].durabilityLoss)
		frame:SetWidth(GameTooltip:GetWidth() - 20)
		GameTooltip_InsertFrame(GameTooltip, frame);
	end
	for i = 1, readyTroops do
		GameTooltip:AddDoubleLine(self.name, READY, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	end
	if ( trainingTroops > 0 ) then
		local now = time()
		local workorder = tfind(MissionMinderData.Characters[CURRENT_PLAYER].followerWorkOrders, self.name, "name")

		for i = 1, trainingTroops do
			--GameTooltip:AddDoubleLine(self.name, WORK_IN_PROGRESS)
			GameTooltip:AddDoubleLine(self.name, secondsToString(workorder.endDate - now + ((i-1)*workorder.duration)), NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 0.7, 0.7, 0.7)
		end
	end
	for i = 1, availableToStart do
		GameTooltip:AddDoubleLine(self.name, L_NOT_STARTED, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	end
	GameTooltip:Show()
end


local function requestCategoryUpdate()
	C_Garrison.RequestClassSpecCategoryInfo(LE_FOLLOWER_TYPE_GARRISON_7_0);
end

local function requestShipmentUpdate()
	C_Garrison.RequestLandingPageShipmentInfo()
end

---------------------------------------------
-- INITIALIZE
---------------------------------------------
function module.initialize()
	-- Update Troop buttons with extended details We must override the original function to avoid
	--rather unsightly code to determine which frame was first (initial order is [1],[2]; but after
	--being released it is [2],[1])
	OrderHallCommandBar.RefreshCategories = displayExtendedTroopDetails

	-- Update Troop tooltip
	hooksecurefunc(OrderHallClassSpecCategory, "OnEnter", showExtendedTroopDetails)


	-- Update categories shipments whenever player starts training new troops
	module:RegisterEvent("SHIPMENT_UPDATE", function(event, updated)
		if ( updated ) then requestShipmentUpdate(); end
	end)

	-- Update categories in response to updated shipments
	module:RegisterEvent("GARRISON_LANDINGPAGE_SHIPMENTS", requestCategoryUpdate)

	-- When a follower is added or removed we must update shipment information so current details
	--are provided for the category refresh. Internally this would cause :RefreshCategories to be
	--called twice. ADD/REMOVE event, :RefreshCategories [1], requestShipmentUpdate,
	--LANDINGPAGE_SHIPMENTS event, :RefreshCategories [2].
	module:RegisterEvent("GARRISON_FOLLOWER_ADDED", requestShipmentUpdate)
	module:RegisterEvent("GARRISON_FOLLOWER_REMOVED", requestShipmentUpdate)
end
