local MAJOR, MINOR = "LibDataUpgrade-1.0", 0
local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end -- Already loaded


local function parseVersion(version)
    local VERSION_REGEX = "^v?(%d+)\.?(%d*)\.?(%d*)$"
    --VERSION_REGEX = "%d+"

    if ( type(version) == "number" ) then
        return version, 0, 0;
    elseif ( type(version) == "string" and version:gmatch(VERSION_REGEX)) then
        local parts = {}
        for _, v in ipairs( {version:match(VERSION_REGEX)} ) do
            if ( v  and v ~= "" ) then
                table.insert(parts, tonumber(v))
            else
                table.insert(parts, 0)
            end
        end
        return unpack(parts)
    else
        error(string.format("Unexpected version format, don't know how to proceed. %s (type %s)", tostring(version), type(version)))
    end
end
--[===[@debug@
local function testParseVersion(version, expected)
    local expectedMajor, expectedMinor, expectedPatch = unpack(expected)
    local major, minor, patch = parseVersion(version)
    assert(major == expectedMajor, string.format("Expected: %s, Actual: %s", expectedMajor, major))
    assert(minor == expectedMinor, string.format("Expected: %s, Actual: %s", expectedMinor, minor))
    assert(patch == expectedPatch, string.format("Expected: %s, Actual: %s", expectedPatch, patch))
end
testParseVersion(1, {1, 0, 0})
testParseVersion("1", {1, 0, 0})
testParseVersion("1.0", {1, 0, 0})
testParseVersion("1.0.3", {1, 0, 3})
testParseVersion("v1", {1, 0, 0})
testParseVersion("v1.1", {1, 1, 0})
testParseVersion("v1.02", {1, 2, 0})
testParseVersion("v1.20", {1, 20, 0})
testParseVersion("1.2.19", {1, 2, 19})
--@end-debug@]===]

local function isLesserVersion(v1, v2)
    local major1, minor1, patch1 = parseVersion(v1)
    local major2, minor2, patch2 = parseVersion(v2)
    if ( major1 ~= major2 ) then
        return major1 < major2
    elseif ( minor1 ~= minor2 ) then
        return minor1 < minor2
    else
        return patch1 < patch2
    end
end

local function versionComparator(o1, o2)
    return isLesserVersion(o1.version, o2.version)
end

--[===[@debug@
assert(isLesserVersion(1, 2) == true)
assert(isLesserVersion("1", "v1.1") == true)
assert(isLesserVersion("1.2", "v1.2.1") == true)
assert(isLesserVersion("1.2", "v1.20") == true)
assert(isLesserVersion("1.02", "v1.20") == true)
assert(isLesserVersion("1.3.3", "v1.3.4") == true)
assert(isLesserVersion("1.3.3", "v1.4.3") == true)
assert(isLesserVersion("1.3.3", "v2.3.3") == true)

assert(isLesserVersion(2, 1) == false)
assert(isLesserVersion("v1.1", "1") == false)
assert(isLesserVersion("1.20", "1.2") == false)
assert(isLesserVersion("1.20", "v1.02") == false)
assert(isLesserVersion("1.3.4", "v1.3.3") == false)
assert(isLesserVersion("1.4.3", "v1.3.3") == false)
assert(isLesserVersion("2.3.3", "v1.3.3") == false)
--@end-debug@]===]

local function deepclone(instance)
    local obj = {}
    for k, v in pairs(instance) do
        if ( type(v) == "table" ) then
            obj[k] = deepclone(v)
        else
            obj[k] = v
        end
    end
    return obj;
end

local function DEBUG(shouldLog, ...)
    if ( shouldLog ) then
        if ( select('#', ...) == 1 and type(select(1,...)) == "table") then
            DevTools_Dump(...)
        else
            print(...)
        end
    end
end

local DATA = {
    Version = function(self, version, f)
        if ( type(f) == "table" ) then
            local base = f
            f = function(table) return deepclone(base) end
        end

        table.insert(self.versions, {
            version = version;
            func = f
        })
    end;

    Defaults = function(self, default)
        return self:Upgrade(default or {})
    end;

    Upgrade = function(self, data, ...)
        table.sort(self.versions, versionComparator)

        data = data or {}
        data[self.versionProperty] = data[self.versionProperty] or 0

        -- upgrade savedVariable
        for _, version in pairs(self.versions) do
            if ( isLesserVersion(data[self.versionProperty], version.version) ) then
                DEBUG(self.debug, "LibDataUpgrade:Load() Upgrading to version", version.version)
                local newdata = version.func(data, ...)
                data = newdata or data -- essentially: if (newdata ~= nil ) then data = newdata; end
                data[self.versionProperty] = version.version
                DEBUG(self.debug, data)
            end
        end

        return data
    end;
}
DATA.__index = DATA



function lib:New(versionProperty)
    local obj = {
        versionProperty = versionProperty or "Version";
        versions = {};

        debug = false;
    }
    setmetatable(obj, DATA)
    return obj;
end
