local ADDON_NAME, ENGINE = ...
-- ENGINE has not been initialized

local ext = {}

local function newCallbackRegistry()
	local registry = {};
	local callbacks = LibStub("CallbackHandler-1.0"):New(registry)

	return {
		Register = function(self, module, event, func)
			if ( type(event) == "string" ) then
				registry.RegisterCallback(module, event, func)
			else
				for i = 1, #event do
					registry.RegisterCallback(module, event[i], func)
				end
			end
		end,

		Fire = function(self, event, ...)
			callbacks:Fire(event, ...);
		end
	}
end


---------------------------------------------
-- Module
---------------------------------------------
local ModuleClass = {
	RegisterEvent = function(self, event, func)
		self.lib.eventFrame:RegisterEvent(event)
		self.lib.callbacks.api:Register(self.fullname, event, func)
	end;

	On = function(self, event, func)
		self.lib.callbacks.events:Register(self.fullname, event, func)
	end;

	Fire = function(self, event, ...)
		ext.DEBUG("[INTERNAL EVENT]", event, ...)
		self.lib.callbacks.events:Fire(event, ...)
	end;
}
ModuleClass.__index = ModuleClass


---------------------------------------------
-- Addon
---------------------------------------------
local AddonClass = {
	SetDefaultOptions = function(self, defaults)
		self.defaultOptions = defaults
	end;

	CreateModule = function(self, name)
		if ( self.modules[name] ) then
			return self.modules[name]
		end

		local module = {
			name = name;
			fullname = self.name .. "/" .. name;
			addon = self;
			lib = self.lib;

			initialize = function() end;
			initialized = false;
		}
		setmetatable(module, ModuleClass)
		self.modules[name] = module

		return module
	end;
}
AddonClass.__index = AddonClass


local function splitstr(str, delimiter)
	local elements = {}
	for i in str:gmatch("[^" .. delimiter .. "]+") do
		table.insert(elements, i)
	end

	return unpack(elements)
end

---------------------------------------------
-- Library
---------------------------------------------
local LibraryClass = {
	SetDefaultOptions = function(self, defaults)
		self.defaultOptions = defaults
	end;

	GetModule = function(self, name)
		local addonName, moduleName = splitstr(name, "/")
		local addon = self.addons[addonName]
		assert( addon ~= nil )
		local module = addon.modules[moduleName]
		assert( module ~= nil )
		return module;
	end;

	CreateAddon = function(self, name)
		if ( self.addons[name] ) then
			return self.addons[name];
		end

		local addon = {
			name = name;
			lib = self;
			version = GetAddOnMetadata(name, "Version");

			modules = {};
		}
		setmetatable(addon, AddonClass)
		self.addons[name] = addon
		return addon
	end;
}
LibraryClass.__index = LibraryClass





---------------------------------------------
-- Initialize
---------------------------------------------
MissionMinder = {
	name = ADDON_NAME;
	ptr = GetBuildInfo():match('7[.]1[.][0-9]+');
	locale = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME);
	eventFrame = CreateFrame("Frame");

	CURRENT_PLAYER = GetUnitName("player") .. "-" .. GetRealmName();

	debug = false;
	initialized = false;

	options = LibStub("LibSavedVariables-1.0"):New("MissionMinderOptions");
	ext = ext;

	addons = {};
	callbacks = {
		api = newCallbackRegistry();
		events = newCallbackRegistry();
	};
}
setmetatable(MissionMinder, LibraryClass)

table.insert(ENGINE, MissionMinder:CreateAddon(ADDON_NAME)) -- addon
table.insert(ENGINE, MissionMinder.options) -- options
table.insert(ENGINE, MissionMinder.locale) -- locale
table.insert(ENGINE, MissionMinder.ext) -- ext

MissionMinder.options:On("Debug", function(option, value) -- Enable / Disable debugging based on option
	MissionMinder.debug = value
end)

MissionMinder.eventFrame:SetScript("OnEvent", function(self, event, ...)
	if (event ~= "ADDON_LOADED") then
		ext.DEBUG("[API]", event, ...)
		MissionMinder.callbacks.api:Fire(event, ...)
	elseif ( event == "ADDON_LOADED" ) then
		local loadedAddon = select(1, ...)
		if ( loadedAddon == ADDON_NAME ) then
			MissionMinder.options:Load()
		end

		local addon = MissionMinder.addons[loadedAddon]
		if ( addon ) then
			ext.foreach(addon.modules, function(_, module)
				module.initialize()
			end)
		end
	end
end)
MissionMinder.eventFrame:RegisterEvent("ADDON_LOADED")
