local addon, options, L, ext = unpack(select(2, ...))
local module = addon:CreateModule("autoClose")

local DEBUG = ext.DEBUG
local DEBUGF = ext.DEBUGF
----------------
-- There are a number of ways the CompleteFrame can initiate AutoClose
-- 1. Mission Failed - Close on failure animation finished and after base XP has been awarded
-- 2. Mission Success, No Bonus XP - Close on chest open animation finished
-- 3. Mission Success, Bonus XP - Close after bonus XP has been awarded
----------------


local AUTOCLOSE_FAILED_DELAY = 1.5
local AUTOCLOSE_SUCCESS_LOOT_DELAY = 1.0;
local AUTOCLOSE_SUCCESS_XP_DELAY = 0.7;

local currentMission = nil;
local hasBonusXP = nil;
local hasBeenLooted = nil;
local autocloseTimer = nil;


local function hideCompleteMission(missionComplete)
	DEBUG("called hideCompleteMission");
	autocloseTimer = nil;
	if ( options.ListViewEnabled() ) then
		missionComplete:GetParent():HideCompleteMissions(true);
	else
		missionComplete.NextMissionButton:Click()
	end
end

local function closeCompletionFrame(missionComplete, delay)
	if( autocloseTimer ) then
		autocloseTimer:Cancel();
	end

	DEBUGF("Closing in %.1f", delay);
	autocloseTimer = C_Timer.NewTimer(delay, function()
		hideCompleteMission(missionComplete)
	end);
end

local function missionHasBonusXPReward(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			return true;
		end
	end

	return false;
end

function MissionMinder_ClearAutoLootState(self)
	DEBUGF("AutoClose::clearState")
	if( autocloseTimer ) then
		autocloseTimer:Cancel();
	end

	currentMission = nil;
	hasBonusXP = nil;
	hasBeenLooted = nil;
	autocloseTimer = nil;

	local followers = self.Stage.FollowersFrame.Followers;
	for i = 1, #followers do
		followers[i].hasFinishedXPGains = nil;
	end
end

function MissionMinder_InitializeAutoCloseState(self, completeMissions, index)
	local missionComplete = self.MissionComplete
	MissionMinder_ClearAutoLootState(missionComplete)
	DEBUGF("AutoClose::initializeState")

	if( options.AutoClose() and missionComplete:IsShown() ) then
		currentMission = completeMissions[index]
		hasBonusXP = missionHasBonusXPReward(currentMission.missionID);
		hasBeenLooted = false;

		local followersRequringXP = #currentMission.followers
		local followers = missionComplete.Stage.FollowersFrame.Followers;
		for i = 1, #currentMission.followers do
			local _, _, level, quality, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, isTroop = C_Garrison.GetFollowerMissionCompleteInfo(currentMission.followers[i]);

			followers[i].hasFinishedXPGains = false;
			if(isTroop or (hasBonusXP and level == self.followerMaxLevel and quality == self.followerMaxQuality)) then
				followers[i].hasFinishedXPGains = true; -- Follower is already maxxed, so ignore XP gains for them
				followersRequringXP = followersRequringXP - 1
			end
			DEBUGF("AutoClose::initState [%d].hasFinishedXPGains = %s", i, tostring(followers[i].hasFinishedXPGains));
		end

		if ( followersRequringXP <= 0 ) then
			hasBonusXP = false
		end
	end

	DEBUGF("AutoClose::initState currentMission = [%d]%s",
		currentMission and currentMission.missionID or 0,
		currentMission and currentMission.name or "nil")
	DEBUGF("AutoClose::initState hasBonusXP = %s", tostring(hasBonusXP))
	DEBUGF("AutoClose::initState hasBeenLooted = %s", tostring(hasBeenLooted))
end



function MissionMinder_MissionComplete_AutoCloseMissionFailed(self)
	local missionComplete = self:GetParent():GetParent()
	if ( options.AutoClose() ) then
		if ( currentMission and not currentMission.succeeded ) then
			print("Mission failed ::: CLOSING");
			closeCompletionFrame(missionComplete, AUTOCLOSE_FAILED_DELAY);
		end
	end
end

function MissionMinder_MissionComplete_MarkMissionLooted()
	hasBeenLooted = true;
end

function MissionMinder_MissionComplete_CloseOnSuccessfulLootNoBonusXP(self)
	local missionComplete = self:GetParent():GetParent():GetParent()
	if ( options.AutoClose() and not hasBonusXP ) then
		print("Mission looted, no bonus XP ::: CLOSING");
		closeCompletionFrame(missionComplete, AUTOCLOSE_SUCCESS_LOOT_DELAY);
	end
end


local function checkFollower(missionComplete, followerFrame)
	if ( not hasBeenLooted ) then
		return false;
	end

	if( followerFrame.activeAnims == nil or followerFrame.activeAnims == 0 ) then
		local pendingXPAwards = missionComplete.pendingXPAwards;
		for k, v in pairs(pendingXPAwards) do
			if ( v.followerID == followerID ) then
				return false; -- has pending rewards
			end
		end

		return true;
	end

	return false;
end

local function isFinishedWithXP(missionComplete)
	local followers = missionComplete.Stage.FollowersFrame.Followers;
	for i = 1, #currentMission.followers do
		if ( not  followers[i].hasFinishedXPGains ) then
			return false;
		end
	end

	return true;
end

function MissionMinder_MissionComplete_CloseOnFinishedXP(self, followerFrame)
	followerFrame.hasFinishedXPGains = checkFollower(self, followerFrame);

	if ( options.AutoClose() and hasBonusXP and isFinishedWithXP(self)) then
		print("Mission looted, with bonus XP ::: CLOSING")
		closeCompletionFrame(self, AUTOCLOSE_SUCCESS_XP_DELAY); -- Maxed out followers never get to this point as they don't have any XP animation
	end
end



function module.register(missionFrame)
	local missionComplete = missionFrame.MissionComplete

--[[
  There are a number of ways the CompleteFrame can initiate AutoClose
   1. Mission Failed - Close on failure animation finished and after base XP has been awarded
   2. Mission Success, No Bonus XP - Close on chest open animation finished
   3. Mission Success, Bonus XP - Close after bonus XP has been awarded
--]]

	-- Mark mission as looted
	module:RegisterEvent("GARRISON_MISSION_BONUS_ROLL_COMPLETE", MissionMinder_MissionComplete_MarkMissionLooted);

	-- Initialize mission state for auto close
	hooksecurefunc(missionFrame, "MissionCompleteInitialize", MissionMinder_InitializeAutoCloseState);
	-- Remove misison state on close
	missionComplete:HookScript("OnHide", MissionMinder_ClearAutoLootState);
	--hooksecurefunc(missionFrame, "HideCompleteMissions", removeMissionState);

	-- Auto close on successful loot
	hooksecurefunc(missionComplete, "OnFollowerXPFinished", MissionMinder_MissionComplete_CloseOnFinishedXP);

	-- Auto close on successful loot (if no bonus XP)
	missionComplete.BonusRewards.ChestModel.OpenAnim:SetScript("OnFinished", MissionMinder_MissionComplete_CloseOnSuccessfulLootNoBonusXP);

	-- Auto close on failure
	missionComplete.ChanceFrame.ResultAnim:HookScript("OnFinished", MissionMinder_MissionComplete_AutoCloseMissionFailed);
end
