local addon, options, L, ext = unpack(select(2, ...))
local module = addon:CreateModule("quickLoot")

local DEBUG = ext.DEBUG
local DEBUGF = ext.DEBUGF


---------------------------------------------
-- LOCALE
---------------------------------------------
local L_MISSION_SUCCESS = L["Mission %s was successful! Received %s"]
local L_MISSION_FAILED = L["Mission %s failed!"]
local L_FOLLOWER_XP_RECEIVED = L["Follower %s received +%d XP."]
local L_FOLLOWER_LEVELUP = L["Follower %s has reached level %d!"]
local L_FOLLOWER_QUALITYUP = L["Follower %s has reached %s quality!"]
local looting = {}



---------------------------------------------
-- Quick loot
---------------------------------------------
function MissionMinder_LootMission(mission)
	if ( mission ) then
		if ( mission.completed ) then
			if ( C_Garrison.CanOpenMissionChest(mission.missionID) ) then
				C_Garrison.MissionBonusRoll(mission.missionID);
			else
				-- Mission cannot be looted; take no action
			end
		else
			looting[mission.missionID] = mission
			C_Garrison.MarkMissionComplete(mission.missionID);
		end
	end
end

local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or "";
end

local function getRewards(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	local str = {};

	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			local xp = BreakUpLargeNumbers(reward.followerXP)
			tinsert(str, "+" .. xp .. " " .. createInlineTexture(reward.icon));
		elseif (reward.itemID) then
			local _, link = GetItemInfo(reward.itemID);
			if(reward.quantity == 1) then
				tinsert(str, link);
			else
				tinsert(str, reward.quantity .. "x " .. link);
			end
		elseif(reward.currencyID) then
			if (reward.currencyID == 0) then -- gold
				tinsert(str, GetCoinTextureString(reward.quantity));
			else
				tinsert(str, reward.quantity .. " " .. createInlineTexture(reward.icon));
			end
		else
			tinsert(str, reward.title);
		end
	end

	return table.concat(str, ", ");
end

local function MissionHasBonusXPReward(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			return true;
		end
	end

	return false;
end

local lootedMissions = {} -- looted missions map is kept as long as inprogress is visible
local function missionSuccessFail(event, missionID, canComplete, succeeded)
	if ( options.QuickLoot() ) then
		if ( lootedMissions[missionID] ) then
			return;
		else
			lootedMissions[missionID] = true
		end

		local missionLink = C_Garrison.GetMissionLink(missionID);
		if ( succeeded ) then
			local rewards = getRewards(missionID);
			SendSystemMessage(string.format(L_MISSION_SUCCESS, missionLink, rewards));
			PlaySound("UI_Garrison_CommandTable_MissionSuccess_Stinger");
		else
			SendSystemMessage(string.format(L_MISSION_FAILED, missionLink));
			PlaySound("UI_Garrison_Mission_Complete_MissionFail_Stinger");
		end

		if ( C_Garrison.CanOpenMissionChest(missionID)) then
			C_Garrison.MissionBonusRoll(missionID);
		end
	end
end

local function clearLootedMissions()
	lootedMissions = {}
end

local xpgains = {};
local looting = {}
local function initializeXPState(event, missionID, canComplete, succeeded)
	local mission = looting[missionID]
	local hasBonusXP = MissionHasBonusXPReward(missionID);

	if ( mission ) then
		looting[missionID] = nil
		for i=1, #(mission.followers) do
			local followerID = mission.followers[i];
			if ( succeeded and hasBonusXP ) then
				xpgains[followerID] = {};
				xpgains[followerID].bonusXP = true;
			elseif(xpgains[followerID]) then
				xpgains[followerID] = nil;
			-- else : if element is aleady nil, do nothing
			end
		end
	end
end

local function reportXPGains(followerID, xpAward, oldXP, oldLevel, oldQuality)
	local follower =  C_Garrison.GetFollowerInfo(followerID);
	local followerLink = C_Garrison.GetFollowerLink(follower.followerID) or follower.name;

	local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(oldLevel, oldQuality);
	if ( maxXP ) then  -- nil if already maxLevel and maxQuality

		if ( options.ShowFollowerXPGains() ) then
			SendSystemMessage(string.format(L_FOLLOWER_XP_RECEIVED, followerLink, xpAward));
		end

		while ( maxXP and xpAward + oldXP >= maxXP ) do -- handle multiple level ups
			xpAward = xpAward - (maxXP - oldXP) -- calculate next level spillover
			oldXP = 0

			if ( oldLevel < GarrisonMissionFrame.followerMaxLevel ) then
				SendSystemMessage(string.format(L_FOLLOWER_LEVELUP, followerLink, oldLevel + 1));
				oldLevel = oldLevel + 1
				maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(oldLevel, oldQuality);
			elseif ( oldQuality < GarrisonMissionFrame.followerMaxQuality ) then
				SendSystemMessage(string.format(L_FOLLOWER_QUALITYUP, followerLink, getglobal("ITEM_QUALITY" .. oldQuality + 1 .. "_DESC")));
				oldQuality = oldQuality + 1
				maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(oldLevel, oldQuality);
			end
		end
	end
end

local function FollowerXPChanged(event, followerID, xpAward, oldXP, oldLevel, oldQuality)
	DEBUG("Follower %s received %d xp", followerID, xpAward);

	local gains = xpgains[followerID];
	if ( gains and gains.bonusXP ) then
		if ( not gains.xpAward ) then
			gains.xpAward = xpAward
			gains.oldXP = oldXP
			gains.oldLevel = oldLevel
			gains.oldQuality = oldQuality

			return -- bonusXP still expected, so we need to aggregate 2 separate GARRISON_FOLLOWER_XP_CHANGED events for each follower
		else
			xpAward = xpAward + gains.xpAward
			oldXP = gains.oldXP
			oldLevel = gains.oldLevel
			oldQuality = gains.oldQuality
		end
	end

	xpgains[followerID] = nil;

	reportXPGains(followerID, xpAward, oldXP, oldLevel, oldQuality)
end


---------------------------------------------
-- INITIALIZE
---------------------------------------------
function module.register(missionFrame)
	module:RegisterEvent("GARRISON_MISSION_COMPLETE_RESPONSE", function(...)
		missionSuccessFail(...)
		--initializeXPState(...)
	end);
	--module:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED", FollowerXPChanged);

	local missionList = missionFrame.MissionTab.MissionList
	missionList:HookScript("OnHide", clearLootedMissions)
	missionList.Tab1:HookScript("OnClick", clearLootedMissions)
end
