local addon, options, L, ext = unpack(select(2, ...))
local module = addon:CreateModule("tracker")

local LibItemInfo = LibStub("LibItemInfo-1.0")


local DEBUG = ext.DEBUG
local DEBUGF = ext.DEBUGF
local secondsToString = ext.secondsToString
local treduce = ext.treduce
local foreach = ext.foreach

---------------------------------------------
-- LOCALE
---------------------------------------------
local COMPLETED = L["completed"]
local L_GARRISON_REPORT_SUMMARY_MISSIONS = GARRISON_MISSIONS
local L_GARRISON_REPORT_SUMMARY_MISSIONS_DESCRIPTION = L["Mission summary for your characters"]
local L_GARRISON_REPORT_SUMMARY_CACHE = GARRISON_CACHE
local L_GARRISON_REPORT_SUMMARY_CACHE_DESCRIPTION = L["Garrison caches that are nearly full"]
local L_GARRISON_REPORT_SUMMARY_BUIDLINGS = GARRISON_LANDING_STATUS_BUILDING
local L_GARRISON_REPORT_SUMMARY_BUIDLINGS_DESCRIPTION = L["Building construction status"]
local L_GARRISON_REPORT_SUMMARY_WORKORDERS = CAPACITANCE_WORK_ORDERS
local L_GARRSION_REPORT_SUMMARY_WORKORDERS_DESCRIPTION = L["Work orders nearing completion"]
local L_GARRISON_REPORT_SUMMARY_FOLLOWER_WORKORDERS = L["Troops training"]
local L_GARRISON_REPORT_SUMMARY_FOLLOWER_WORKORDERS_DESCRIPTION = L["Troops currently being trained"]
local L_GARRISON_REPORT_SUMMARY_TALENTS = L["Research"]
local L_GARRISON_REPORT_SUMMARY_TALENTS_DESCRIPTION = L["OrderHall talents being researched"]
local L_TROOPS_READY = L["Troops Ready"];
local L_WORKORDER_COMPLETED = L["WorkOrder Completed"];
local L_RESEARCH_COMPELTED = L["Research Completed"];

local L_GARRISON_TYPE = GARRISON_LOCATION_TOOLTIP
local L_ORDERHALL_TYPE = L["OrderHall"]

---------------------------------------------
-- DataTypes
---------------------------------------------
local CURRENT_PLAYER = GetUnitName("player") .. "-" .. GetRealmName()

function MissionData(mission)
	local data = {
		_type = "mission";
		_player = CURRENT_PLAYER;
		name = mission.name;
		missionID = mission.missionID;
		level = mission.level;
		iLevel = mission.iLevel;
		type = mission.type;
		typeAtlas = mission.typeAtlas;
		location = mission.location;
		locPrefix = mission.locPrefix;
		isRare = mission.isRare;

		numFollowers = mission.numFollowers;
		followers = {};
		numRewards = mission.numRewards;
		rewards = {};
	}

	if ( mission.followers ) then
		for i = 1, #mission.followers do
			local followerID = mission.followers[i]
			tinsert(data.followers, C_Garrison.GetFollowerName(followerID))
		end

		if ( #mission.followers > 0 ) then
			local timeleft = C_Garrison.GetFollowerMissionTimeLeftSeconds(mission.followers[1]);
			if ( timeleft ) then
				data.endDate = time() + timeleft
			end
		end
	end

	for k,reward in pairs(mission.rewards) do
		if ( reward.quality ) then
			tinsert(data.rewards, ITEM_QUALITY_COLORS[reward.quality + 1].hex..reward.title..FONT_COLOR_CODE_CLOSE)
		elseif ( reward.itemID ) then
			-- Attempt to get reward details,
			LibItemInfo:GetItemInfo(reward.itemID, function(itemName, _, itemRarity)
				tinsert(data.rewards, ITEM_QUALITY_COLORS[itemRarity].hex..itemName..FONT_COLOR_CODE_CLOSE);
			end)
			--[[local itemName, _, itemRarity = GetItemInfo(reward.itemID);
			if ( itemName ) then
				tinsert(data.rewards, ITEM_QUALITY_COLORS[itemRarity].hex..itemName..FONT_COLOR_CODE_CLOSE);
			end]]
		else
			tinsert(data.rewards, reward.title)
		end
	end

	return data
end

local GARRISON_TYPES = {
	[LE_GARRISON_TYPE_6_0] = L_GARRISON_TYPE;
	[LE_GARRISON_TYPE_7_0] = L_ORDERHALL_TYPE;
}



---------------------------------------------
--
---------------------------------------------
local GARRISON_CURRENCY = 824


local SECONDS, MINUTES, HOURS, DAYS = unpack(ext.TIMEUNITS)


local MAX_GARRISON_CACHE = 500
local GARRISON_CACHE_RATE_IN_SECONDS = 10 * MINUTES
local GARRISON_CACHE_WARN_AMOUNT = MAX_GARRISON_CACHE - (1 * DAYS / GARRISON_CACHE_RATE_IN_SECONDS)

local GARRISON_RESOURCES_TEXTURE=select(3, GetCurrencyInfo(GARRISON_CURRENCY))


local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or "";
end


function module.estimateGarrisonCache(player)
	local now = time()
	local lastLooted = MissionMinderData.Characters[player].resourceCacheLastLooted

	if ( lastLooted ) then
		return min(MAX_GARRISON_CACHE, floor((now - lastLooted) / GARRISON_CACHE_RATE_IN_SECONDS))
	else
		return nil
	end
end

--[[function module.estimateGarrisonCacheTimeRemaining(player)
	local now = time()
	local lastLooted = Lib.Globals.Characters[player].GarrisonCache

	if ( lastLooted ) then
		local maxedAt =  lastLooted + (MAX_GARRISON_CACHE * GARRISON_CACHE_RATE_IN_SECONDS)
		return  (maxedAt > now ) and maxedAt - now or 0;
	else
		return nil
	end
end]]

local function orderGarrisonCache(cache1, cache2)
	if ( cache1.amount ~= cache2.amount ) then
		return cache1.amount > cache2.amount
	else
		return strcmputf8i(cache1.name, cache1.name) < 0
	end
end

local function orderTalents(talent1, talent2)
	if ( talent1.garrType ~= talent2.garrType ) then
		return talent1.garrType > talent2.garrType
	elseif ( talent1.player ~= talent2.player ) then
		return strcmputf8i(talent1.player, talent2.player) < 0
	elseif ( talent1.isComplete ~= talent2.isComplete ) then
		return talent1.isComplete
	elseif ( talent1.endDate ~= talent2.endDate ) then
		return talent1.endDate < talent2.endDate
	else
		return strcmputf8i(talent1.name, talent2.name) < 0
	end
end

local function orderBuildings(building1, building2)
	if ( building1.garrType ~= building2.garrType ) then
		return building1.garrType > building2.garrType
	elseif ( building1.player ~= building2.player ) then
		return strcmputf8i(building1.player, building2.player) < 0
	elseif ( building1.isComplete ~= building2.isComplete ) then
		return building1.isComplete
	elseif ( building1.endDate ~= building2.endDate ) then
		return building1.endDate < building2.endDate
	else
		return strcmputf8i(building1.name, building2.name) < 0
	end
end

local function orderWorkOrders(workorder1, workorder2)
	local pending1 = workorder1.total - workorder1.ready
	local pending2 = workorder2.total - workorder2.ready

	if ( workorder1.garrType ~= workorder2.garrType ) then
		return workorder1.garrType > workorder2.garrType
	elseif ( workorder1.player ~= workorder2.player ) then
		return strcmputf8i(workorder1.player, workorder2.player) < 0
	elseif ( pending1 ~= pending2 ) then
		return pending1 < pending2
	else
		return strcmputf8i(workorder1.name, workorder2.name) < 0
	end
end

local function orderTrackedMissions(track1, track2)
	if ( track1.garrType ~= track2.garrType ) then
		return track1.garrType > track2.garrType
	elseif ( track1.completed ~= track2.completed ) then
		return track1.completed > track2.completed
	elseif ( track1.lowest ~= track2.lowest ) then
		return track1.lowest < track2.lowest
	else
		return strcmputf8i(track1.player, track2.player) < 0;
	end
end


local tooltipPool = {}
local activeTooltips = 0
MissionMinder.tooltips = tooltipPool

local function acquireTooltip()
	activeTooltips = activeTooltips + 1
	if ( #tooltipPool < activeTooltips ) then
		local tooltip = CreateFrame("GameTooltip", "MissionMinderReportSummaryTooltip" .. #tooltipPool, nil, "MissionMinderReportSummaryTooltip")
		table.insert(tooltipPool, tooltip)
		return tooltip
	end
	return tooltipPool[activeTooltips]
end

local function hideReportTooltip(self)
	activeTooltips = 0
	for i=1, #tooltipPool do
		tooltipPool[i]:Hide()
		tooltipPool[i]:ClearAllPoints()
	end
end

local function showReportTooltip(self)
	hideReportTooltip(self) -- Make sure we reset the state
	local now = time();

	local insets = GameTooltip:GetBackdrop().insets
	local anchor_y = 4 - insets.top/2 - insets.bottom/2
	local tooltips = { GameTooltip }

	-- Mission progress should always show when available
	local tracking = {}
	for player, character in pairs(MissionMinderData.Characters) do
		for garrType, _ in pairs(GARRISON_TYPES) do
			if ( not options.ShowMissionsSummary(garrType) ) then
				break
			end

			local data = {
				player = player;
				garrType = garrType;
				lowest = nil;
				completed = 0;
			}

			for missionID, mission in pairs(character.missions) do
				if ( mission.garrType == garrType ) then
					if ( mission.endDate <= now ) then
						data.lowest = 0;
						data.completed = data.completed + 1
					elseif ( not data.lowest or mission.endDate - now < data.lowest ) then
						data.lowest = mission.endDate - now;
					end
				end
			end

			if ( data.lowest ) then
				table.insert(tracking, data)
			end
		end
	end
	table.sort(tracking, orderTrackedMissions);
	if ( #tracking > 0 ) then
		local tooltip = acquireTooltip()
		table.insert(tooltips, tooltip)
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips-1], "BOTTOMLEFT", 0, -anchor_y)

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_MISSIONS, 1, 1, 1)
		tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_MISSIONS_DESCRIPTION)
		local prevGarrType = nil
		for i = 1, #tracking do
			if ( prevGarrType ~= tracking[i].garrType ) then
				tooltip:AddLine(" ")
				tooltip:AddLine(GARRISON_TYPES[tracking[i].garrType])
				prevGarrType = tracking[i].garrType
			end

			if ( tracking[i].completed > 0 ) then
				tooltip:AddDoubleLine(tracking[i].player, tracking[i].completed .. " " .. COMPLETED, 1, 1, 1, 0, 1, 0);
			else
				tooltip:AddDoubleLine(tracking[i].player, secondsToString(tracking[i].lowest), 1, 1, 1, 0.7, 0.7, 0.7);
			end
		end
	end

	-- Talent progress should always show when applicable
	local talents = {}
	for player, character in pairs(MissionMinderData.Characters) do
		for  _, talent in pairs(character.talents) do
			if ( options.ShowTalentsSummary(talent.garrType) ) then
				table.insert(talents, {
					player = player,
					garrType = talent.garrType;
					name = talent.name;
					isComplete = (now >= talent.endDate);
					endDate = talent.endDate;
				})
			end
		end
	end
	table.sort(talents, orderTalents)
	if ( #talents > 0 ) then
		local tooltip = acquireTooltip()
		table.insert(tooltips, tooltip)
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips-1], "BOTTOMLEFT", 0, -anchor_y)

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_TALENTS, 1, 1, 1)
		tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_TALENTS_DESCRIPTION)

        tooltip:AddLine(" ")
		for i = 1, #talents do
			if ( talents[i].isComplete ) then
				tooltip:AddDoubleLine(talents[i].player, COMPLETED, 1, 1, 1, 0, 1, 0);
			else
				tooltip:AddDoubleLine(talents[i].player, secondsToString(talents[i].endDate - now), 1, 1, 1, 0.7, 0.7, 0.7)
			end
		end
	end

	-- Building progress should always show when applicable
	local buildings = {}
	for player, character in pairs(MissionMinderData.Characters) do
		for  _, building in pairs(character.buildings) do
			if ( options.ShowBuildingsSummary(building.garrType) ) then
				if ( building.isBuilding or building.canActivate ) then
					tinsert(buildings,{
						player = player,
						garrType = building.garrType;
						name = building.name;
						isComplete = (now >= building.endDate);
						endDate = building.endDate;
					})
				end
			end
		end
	end
	table.sort(buildings, orderBuildings)
	if ( #buildings > 0 ) then
		local tooltip = acquireTooltip()
		tinsert(tooltips, tooltip)
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips-1], "BOTTOMLEFT", 0, -anchor_y)

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_BUIDLINGS, 1, 1, 1)
		tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_BUIDLINGS_DESCRIPTION)

		local lastPlayer = nil
		for i = 1, #buildings do
			if ( lastPlayer ~= buildings[i].player ) then
				tooltip:AddLine(" ")
				tooltip:AddLine(buildings[i].player)
				lastPlayer = buildings[i].player
			end

			if ( buildings[i].isComplete ) then
				tooltip:AddDoubleLine(buildings[i].name, COMPLETED, 1, 1, 1, 0, 1, 0);
			else
				tooltip:AddDoubleLine(buildings[i].name, secondsToString(buildings[i].endDate - now), 1, 1, 1, 0.7, 0.7, 0.7)
			end
		end
	end

	-- if any GarrisonCache is > 356 (less than 1 day from maxing)
	if ( options.ShowGarrisonCacheSummary() ) then
		local garrisonCaches = {};
		for player, _ in pairs(MissionMinderData.Characters) do
			local cache = module.estimateGarrisonCache(player)
			if ( cache and cache >= GARRISON_CACHE_WARN_AMOUNT ) then
				tinsert(garrisonCaches, {
					name = player,
					amount = cache
				});
			end
		end
		table.sort(garrisonCaches, orderGarrisonCache)
		if ( #garrisonCaches > 0 ) then
			local tooltip = acquireTooltip()
			tinsert(tooltips, tooltip)
			tooltip:SetOwner(self, "ANCHOR_NONE")
			tooltip:SetPoint("TOPLEFT", tooltips[#tooltips-1], "BOTTOMLEFT", 0, -anchor_y)

			tooltip:SetText(L_GARRISON_REPORT_SUMMARY_CACHE, 1, 1, 1)
			tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_CACHE_DESCRIPTION)
			tooltip:AddLine(" ")
			for i = 1, #garrisonCaches do
				tooltip:AddDoubleLine(garrisonCaches[i].name, garrisonCaches[i].amount .. " " .. createInlineTexture(GARRISON_RESOURCES_TEXTURE), 1, 1, 1, 1, 0, 0);
			end
		end
	end

	-- if any work orders are nearly complete and need to be refilled in less than a day
	local follower = {}
	for player, character in pairs(MissionMinderData.Characters) do
		for _, workorder in pairs(character.followerWorkOrders) do
			if ( options.ShowTroopsSummary(workorder.garrType) ) then
				local name = workorder.name
				local shipmentCapacity = workorder.shipmentCapacity
				local shipmentsReady = workorder.shipmentsReady
				local shipmentsTotal = workorder.shipmentsTotal
				local creationTime = workorder.creationTime
				local duration = workorder.duration
				local endDate = workorder.endDate

				if ( name and shipmentCapacity and shipmentCapacity > 0  and shipmentsTotal ) then
					if ( duration > 0 and now > creationTime + duration ) then -- Adjust ready based on elapsed time
						shipmentsReady = min(shipmentsTotal, shipmentsReady + (now - creationTime) / duration)
					end

					local timeRemaining = (shipmentsTotal - shipmentsReady) * duration
					if ( timeRemaining < 1 * DAYS ) then
						tinsert(follower, {
							player = player;
							garrType = workorder.garrType;
							name = name;
							ready = floor(shipmentsReady);
							total = shipmentsTotal;
							endDate = endDate;
						});
					end
				end
			end
		end
	end
	sort(follower, orderWorkOrders)
	if ( #follower > 0 ) then
		local tooltip = acquireTooltip()
		tinsert(tooltips, tooltip)
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips-1], "BOTTOMLEFT", 0, -anchor_y)

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_FOLLOWER_WORKORDERS, 1, 1, 1)
		tooltip:AddLine(L_GARRISON_REPORT_SUMMARY_FOLLOWER_WORKORDERS_DESCRIPTION)

		local lastPlayer = nil
		for i = 1, #follower do
			if ( lastPlayer ~= follower[i].player ) then
				tooltip:AddLine(" ")
				tooltip:AddLine(follower[i].player)
				lastPlayer = follower[i].player
			end

			if ( follower[i].ready > 0 ) then
				tooltip:AddDoubleLine(follower[i].name, follower[i].ready .. "/" .. follower[i].total, 1, 1, 1, 0, 1, 0);
			else
				tooltip:AddDoubleLine(follower[i].name, secondsToString(follower[i].endDate - now), 1, 1, 1, 0.7, 0.7, 0.7)
			end
		end
	end

	local workorders = {}
	for player, character in pairs(MissionMinderData.Characters) do
		for _, workorder in pairs(character.looseWorkOrders) do
			if ( options.ShowWorkOrdersSummary(workorder.garrType) ) then
				local name = workorder.name
				local shipmentCapacity = workorder.shipmentCapacity
				local shipmentsReady = workorder.shipmentsReady
				local shipmentsTotal = workorder.shipmentsTotal
				local creationTime = workorder.creationTime
				local duration = workorder.duration
				local endDate = workorder.endDate

				if ( name and shipmentCapacity and shipmentCapacity > 0  and shipmentsTotal ) then
					if ( duration > 0 and now > creationTime + duration ) then -- Adjust ready based on elapsed time
						shipmentsReady = min(shipmentsTotal, shipmentsReady + (now - creationTime) / duration)
					end

					tinsert(workorders, {
						player = player;
						garrType = garrType;
						name = name;
						ready = floor(shipmentsReady);
						total = shipmentsTotal;
						endDate = endDate;
					});
				end
			end
		end
	end
	sort(workorders, orderWorkOrders)
	if ( #workorders > 0 ) then
		local tooltip = acquireTooltip()
		tinsert(tooltips, tooltip)
		tooltip:SetOwner(self, "ANCHOR_NONE")
		tooltip:SetPoint("TOPLEFT", tooltips[#tooltips-1], "BOTTOMLEFT", 0, -anchor_y)

		tooltip:SetText(L_GARRISON_REPORT_SUMMARY_WORKORDERS, 1, 1, 1)
		tooltip:AddLine(L_GARRSION_REPORT_SUMMARY_WORKORDERS_DESCRIPTION)

		local lastPlayer = nil
		for i = 1, #workorders do
			if ( lastPlayer ~= workorders[i].player ) then
				tooltip:AddLine(" ")
				tooltip:AddLine(workorders[i].player)
				lastPlayer = workorders[i].player
			end

			if ( workorders[i].ready > 0 ) then
				tooltip:AddDoubleLine(workorders[i].name, workorders[i].ready .. "/" .. workorders[i].total, 1, 1, 1, 0, 1, 0);
			else
				tooltip:AddDoubleLine(workorders[i].name, secondsToString(workorders[i].endDate - now), 1, 1, 1, 0.7, 0.7, 0.7)
			end
		end
	end


	-- Update tooltip width and right anchors
	local largestWidth = 0
	for i = 1, #tooltips do
		local tooltip = tooltips[i]
		tooltip:Show()
		largestWidth = max(largestWidth, tooltip:GetWidth())
	end
	for i = 1, #tooltips do
		local tooltip = tooltips[i]
		local difference = largestWidth - tooltip:GetWidth()
		tooltip:SetWidth(largestWidth)

		for line = 1, 100 do
			local left = _G[tooltip:GetName() .. "TextLeft" .. line]
			if (not left or not left:IsShown()) then -- No more work
				break;
			end

			local right = _G[tooltip:GetName() .. "TextRight" .. line]
			if ( right:IsShown() ) then
				local anchor, relobj, relanchor, x, y = right:GetPoint(1);
				right:SetPoint(anchor, relobj, relanchor, x + difference, y);
			end
		end
	end
end

--[[local function trackGarrisonCache(event, quanity)
	Lib.Globals.Characters[CURRENT_PLAYER].GarrisonCache = time()
end]]

local function trackBuilding(garrType, plotID)
	if ( not plotID ) then
		return;
	end

	local id, name, texPrefix, icon, rank, isBuilding, timeStart, buildTime, canActivate, canUpgrade, isPrebuilt = C_Garrison.GetOwnedBuildingInfoAbbrev(plotID);

	if ( not name ) then
		return
	end

	if ( isBuilding or canActivate ) then -- building is constructing, add to workorders if not currently listed
		local building = {
				_type = "building";
				_player = CURRENT_PLAYER;
				plotID = plotID;
				buildingID = id;
				name = name;
				texPrefix = texPrefix;
				icon = icon;
				rank = rank;
				isBuilding = isBuilding;
				timeStart = timeStart;
				buildTime = buildTime;
				endDate = timeStart + buildTime;
				canActivate = canActivate;
				canUpgrade = canUpgrade;
				isPrebuilt = isPrebuilt;
			}
		building.garrType = garrType
		MissionMinderData.Characters[CURRENT_PLAYER].buildings[plotID] = building
	end
end

function trackFollowerWorkOrder(garrType, containerID)
	local name, texture, shipmentCapacity, shipmentsReady, shipmentsTotal, creationTime, duration, timeleftString, _, _, _, _, followerID = C_Garrison.GetLandingPageShipmentInfoByContainerID(containerID);

	if ( not name ) then
		return;
	end

	local workorder = {
		_type = "followerworkorder";
		_player = CURRENT_PLAYER;
		texture = texture;
		name = name;
		followerID = followerID;
		creationTime = creationTime;
		duration = duration;
		endDate = creationTime + duration;
		shipmentCapacity = shipmentCapacity;
		shipmentsReady = shipmentsReady;
		shipmentsTotal = shipmentsTotal;
	}
	workorder.garrType = garrType
	table.insert(MissionMinderData.Characters[CURRENT_PLAYER].followerWorkOrders, workorder)
end

local function trackLooseWorkOrders(garrType, containerID)
	local name, texture, shipmentCapacity, shipmentsReady, shipmentsTotal, creationTime, duration, timeleftString = C_Garrison.GetLandingPageShipmentInfoByContainerID(containerID);

	if ( not name ) then
		return
	end

	local workorder = {
		_type = "looseworkorder";
		_player = CURRENT_PLAYER;
		texture = texture;
		name = name;
		--buildingID = building.buildingID;
		--plotID = building.plotID;
		creationTime = creationTime;
		duration = duration;
		endDate = creationTime + duration;
		--itemName = itemName;
		shipmentCapacity = shipmentCapacity;
		shipmentsReady = shipmentsReady;
		shipmentsTotal = shipmentsTotal;
	}
	workorder.garrType = garrType
	table.insert(MissionMinderData.Characters[CURRENT_PLAYER].looseWorkOrders, workorder)
end

function trackTalent(garrType, talent)
	if ( talent.isBeingResearched or talent.id == C_Garrison.GetCompleteTalent(garrType)) then
		local endDate = time() + talent.researchTimeRemaining
		local workorder = {
			_type = "talent";
			_player = CURRENT_PLAYER;
			texture = talent.icon;
			name = talent.name;
			description = talent.description;
			creationTime = endDate - talent.researchDuration;
			duration = talent.researchDuration;
			endDate = endDate;
			isBeingResearched = talent.isBeingResearched;
		}
		workorder.garrType = garrType
		table.insert(MissionMinderData.Characters[CURRENT_PLAYER].talents, workorder)
	end
end

--[[
local function trackWorkOrder(event, plotID, maxShipments)
	local numPending = C_Garrison.GetNumPendingShipments();
	if (not numPending) then
		return;
	end

	if ( C_Garrison.IsOnShipmentQuestForNPC() ) then
		maxShipments = 1;
	end

	--local available = maxShipments - numPending;

	-- Update only the workorder for this specific building
	local buildingID, name = C_Garrison.GetOwnedBuildingInfoAbbrev(plotID);
	local workorders = Lib.Globals.Characters[CURRENT_PLAYER].workorders or {}
	for i = 1, #workorders do
		local workorder = workorders[i]
		if ( workorder.buildingID == buildingID and workorder.duration and workorder.duration > 0 ) then
			if ( workorder.shipmentCapacity ~= maxShipments or workorder.shipmentsTotal ~= numPending) then
				DEBUG("Updating building %d \"%s\" workorder: %d / %d", buildingID, name, numPending, maxShipments)
			end

			workorder.shipmentCapacity = maxShipments
			workorder.shipmentsTsotal = numPending
			break;
		end
	end
end

local delayedRushOrderUpdate = nil
local function updateForRushOrder(event, spellID)
	if ( delayedRushOrderUpdate ) then
		delayedRushOrderUpdate:Cancel()
		delayedRushOrderUpdate = nil
	end

	delayedRushOrderUpdate = C_Timer.NewTimer(1 * SECONDS, C_Garrison.RequestLandingPageShipmentInfo);
end]]


function trackAllShipments(self)
	-- Clean any data as we will be pulling all new data for ALL garrTypes
	MissionMinderData.Characters[CURRENT_PLAYER].buildings = {}
	MissionMinderData.Characters[CURRENT_PLAYER].looseWorkOrders = {}
	MissionMinderData.Characters[CURRENT_PLAYER].followerWorkOrders = {}
	MissionMinderData.Characters[CURRENT_PLAYER].talents = {}

	for garrType, _ in pairs(GARRISON_TYPES) do
		local buildings = C_Garrison.GetBuildings(garrType);
		if ( buildings ) then
			for i = 1, #buildings do
				if ( buildings[i].buildingID ) then
					trackBuilding(garrType, buildings[i].plotID)
					--[[local name, texture, shipmentCapacity, shipmentsReady, shipmentsTotal, creationTime, duration, timeleftString, itemName, itemIcon, itemQuality, itemID = C_Garrison.GetLandingPageShipmentInfo(buildingID);
					if ( name and shipmentCapacity > 0 and shipmentIndex < maxShipments ) then
						local shipment = self.shipmentsPool:Acquire();
						SetupShipment(shipment, texture, true, name, buildingID, buildings[i].plotID, nil, shipmentsReady, shipmentsTotal, creationTime, duration, SHIPMENT_TYPE_BUILDING, shipmentIndex);
						shipmentIndex = shipmentIndex + 1;
					end]]
				end
			end
		end

		local followerShipments = C_Garrison.GetFollowerShipments(garrType);
		if ( followerShipments ) then
			for i = 1, #followerShipments do
				trackFollowerWorkOrder(garrType, followerShipments[i])
				--[[local name, texture, shipmentCapacity, shipmentsReady, shipmentsTotal, creationTime, duration, timeleftString, _, _, _, _, followerID = C_Garrison.GetLandingPageShipmentInfoByContainerID(followerShipments[i]);
				if ( name and shipmentCapacity > 0 and shipmentIndex < maxShipments ) then
					local shipment = self.shipmentsPool:Acquire();
					SetupShipment(shipment, texture, false, name, nil, nil, followerShipments[i], shipmentsReady, shipmentsTotal, creationTime, duration, SHIPMENT_TYPE_FOLLOWER, shipmentIndex);
					shipmentIndex = shipmentIndex + 1;
				end]]
			end
		end


		local looseShipments = C_Garrison.GetLooseShipments(garrType);
		if ( looseShipments ) then
			for i = 1, #looseShipments do
				trackLooseWorkOrders(garrType, looseShipments[i])
				--[[local name, texture, shipmentCapacity, shipmentsReady, shipmentsTotal, creationTime, duration, timeleftString = C_Garrison.GetLandingPageShipmentInfoByContainerID(looseShipments[i]);
				if ( name and shipmentCapacity > 0 and shipmentIndex < maxShipments ) then
					local shipment = self.shipmentsPool:Acquire();
					SetupShipment(shipment, texture, true, name, nil, nil, looseShipments[i], shipmentsReady, shipmentsTotal, creationTime, duration, SHIPMENT_TYPE_LOOSE, shipmentIndex);
					shipmentIndex = shipmentIndex + 1;
				end]]
			end
		end

		local talentTrees = C_Garrison.GetTalentTrees(garrType, select(3, UnitClass("player")) --[[classID]]);
		if (talentTrees) then
			for _, tree in ipairs(talentTrees) do
				for _, talent in ipairs(tree) do
					trackTalent(garrType, talent)
					--[[if (showTalent) then
						local shipment = self.shipmentsPool:Acquire();
						SetupShipment(shipment, talent.icon, true, talent.name, nil, nil, nil, talent.isBeingResearched and 0 or 1, 1, talent.researchStartTime, talent.researchDuration, SHIPMENT_TYPE_TALENT, shipmentIndex);
						shipment.talent = talent;
						shipmentIndex = shipmentIndex + 1;
					end]]
				end
			end
		end
	end
end


local function adjustAlertText(self)
	if ( self.CharacterName:IsShown() ) then
		self.AlertText:ClearAllPoints()
		self.AlertText:SetPoint("TOPRIGHT", self, "TOPLEFT", 4, 5);
	else
		self.AlertText:ClearAllPoints()
		self.AlertText:SetPoint("RIGHT", self, "LEFT", 4, 0);
	end
end

local function hideCharacterName()
	local self = GarrisonLandingPageMinimapButton
	self.CharacterName:SetText(nil)
	self.CharacterName:Hide()
	adjustAlertText(self)
end

local function showAlert(alertinfo)
	local self = GarrisonLandingPageMinimapButton
	if ( self:IsShown() ) then
		self.AlertText:SetText(alertinfo.msg);
		GarrisonMinimap_Justify(self.AlertText);

		if ( alertinfo.player ~= CURRENT_PLAYER ) then
			self.CharacterName:SetText(alertinfo.player)
			self.CharacterName:Show()
			adjustAlertText(self)
		end

		self.MinimapAlertAnim:Play();
		if ( alertinfo.sound ) then
			PlaySound(alertinfo.sound);
		end
	end
end

local function findLowestPending(tbl, now)
	now = now or time()

	local lowest, info
	if ( tbl ) then
		for _, data in pairs(tbl) do
			if ( data.endDate > now ) then -- pending
				if ( not lowest or ( data.endDate - now < lowest )) then
					lowest, info = data.endDate - now, data
				end
			end
		end
	end

	return lowest, info
end

local function lowestEndDate(obj1, obj2)
	-- We assume both may be null
	if ( not obj1 ) then
		return obj2
	elseif ( not obj2 ) then
		return obj1
	elseif ( obj1.endDate < obj2.endDate ) then
		return obj1
	else
		return obj2
	end
end

local function generateAlertMsg(info)
	local case = {
		building = string.format(GARRISON_BUILDING_COMPLETE_TOAST, info.name);
		mission = GARRISON_MISSION_COMPLETE;
		followerworkorder = L_TROOPS_READY;
		looseworkorder = L_WORKORDER_COMPLETED;
		talent = L_RESEARCH_COMPELTED;
	}
	return case[info._type]
end

local function generateAlertSound(info)
	local case = {
		building = "UI_Garrison_Toast_BuildingComplete";
		mission = "UI_Garrison_Toast_MissionComplete";
		followerworkorder = nil;
		looseworkorder = nil;
		talent = nil;
	}
	return case[info._type]
end

local function createAlertTimer()
	local nextAlertDelay = nil;
	local alertInfo = nil;
	local alertCharacter = nil

	--[[ Nested treduce, a bit ugly.

		We're essentially looking across all characters, across all tracked details, to
		identify the next item that will complete.
	]]
	local now = time()
	local nextAlertData = treduce(MissionMinderData.Characters, function(player, character, nextAlertData)
		local tables
		if ( player == CURRENT_PLAYER ) then
			-- Missions and builings already have their own popups when finished, we'll important ones
			tables = { character.talents, character.looseWorkOrders, character.followerWorkOrders }
		else
			tables = { character.missions, character.buildings, character.talents, character.looseWorkOrders, character.followerWorkOrders }
		end

		local lowestForCharacter = treduce(tables, function(_, table, lowestForCharacter)
				local lowestInCollection = treduce(table, function(key, value, reduced)-- findLowestPending(tables[i], now)
					if ( value.endDate > now ) then -- not yet finished
						return lowestEndDate(reduced, value)
					end
					return reduced;
				end)
				return lowestEndDate(lowestForCharacter, lowestInCollection)
		end)

		-- Compare using a custom alert structure
		if ( lowestForCharacter ) then
			return lowestEndDate(nextAlertData, {
				player = lowestForCharacter._player;
				type = lowestForCharacter._type;
				name = lowestForCharacter.name;
				endDate = lowestForCharacter.endDate;
				msg = generateAlertMsg(lowestForCharacter);
				sound = generateAlertSound(lowestForCharacter);
			})
		end
		return nextAlertData
	end)
	if ( nextAlertData ) then
		local delay = nextAlertData.endDate - now
		DEBUGF("createAlertTimer(%ds, %s, %s)", delay, nextAlertData.player, nextAlertData.msg)
		C_Timer.After(delay+2, function()
			showAlert(nextAlertData)
			createAlertTimer() -- Create the next timer
		end)
	else
		DEBUG("No more events to alert for")
	end
end

--[[
local function updateGarrisonCurrency()
	if ( not MissionMinder ) then -- Only do this while MissionMinder has not been loaded (once loaded this will be handled by the core addon)
		local _, quantity, _, _, _, maxQuantity = GetCurrencyInfo(GARRISON_CURRENCY);
		DEBUG("updateGarrisonCurrency() for %s = %d/%d", CURRENT_PLAYER, quantity, maxQuantity)
		Lib.Globals.Characters[CURRENT_PLAYER].GarrisonResources = quantity
		Lib.Globals.Characters[CURRENT_PLAYER].MaxGarrisonResources = maxQuantity
	end
end]]


local function trackInProgressMissions(self)
	local now = time()

	local trackedMissions = MissionMinderData.Characters[CURRENT_PLAYER].missions
	--[[if ( not trackedMissions ) then
		trackedMissions = {}
		MissionMinderData.Characters[CURRENT_PLAYER].missions = trackedMissions
	end]]

	local inprogress = self.inProgressMissions
	if ( inprogress ) then
		-- Remove all missions for the given garrType
		local garrType = self:GetParent():GetParent().garrTypeID
		for missionID, mission in pairs(trackedMissions) do
			if ( mission.garrType == garrType ) then
				trackedMissions[missionID] = nil
			end
		end

		-- Load all inprogress missions
		for i = 1, #inprogress do
			local mission = inprogress[i]
			trackedMissions[mission.missionID] = MissionData(mission)
			trackedMissions[mission.missionID].garrType = garrType
		end
	end
end



local data = LibStub("LibDataUpgrade-1.0"):New()
data:Version("1", {
	Characters = {};
})

local chardata = LibStub("LibDataUpgrade-1.0"):New()
chardata:Version("1", {
	resourceCacheLastLooted = nil;
	buildings = {};
	workOrders = {};
	talents = {};
	missions = {};
	currency = {
		--GARRISON_RESOURCE_ID = 0;
		--ORDERHALL_RESOURCE_ID = ;
	};
})
chardata:Version("1.0.1", function(character)
	-- Split follower and loose workOrders into separate entries
	-- This was done early in development, so no data conversion is necessary
	if ( character.workOrders ) then
		character.workOrders = nil
		character.looseWorkOrders = {}
		character.followerWorkOrders = {}
	end

	-- Talents is a separate from buildings or workOrders
	-- This was done early in development, so no data conversion is necessary
	if ( not character.talents ) then
		character.talents = {}
	end
end)
chardata:Version("1.0.2", function(character, player)
	-- Upgrade any existing data to include new _type and _player attributes
	character._player = player
	foreach(character.missions, function(k, v)
		v._type = "mission"
		v._player = player
	end)
	foreach(character.buildings, function(k, v)
		v._type = "building"
		v._player = player
	end)
	foreach(character.talents, function(k, v)
		v._type = "talent"
		v._player = player
	end)
	foreach(character.looseWorkOrders, function(k, v)
		v._type = "looseworkorder"
		v._player = player
	end)
	foreach(character.followerWorkOrders, function(k, v)
		v._type ="followerworkorder"
		v._player = player
	end)
end)

function module.initialize()
	MissionMinderData = data:Upgrade(MissionMinderData) -- Upgrade the base obj
	MissionMinderData.Characters[CURRENT_PLAYER] = chardata:Upgrade(MissionMinderData.Characters[CURRENT_PLAYER]) -- Upgrade the current player (incidentally this will create if nil)
	for player, character in pairs(MissionMinderData.Characters) do -- Upgrade all other character data
		MissionMinderData.Characters[player] = chardata:Upgrade(character, player)
	end


	-- Display mission status on GarrisonReport tooltip
	GarrisonLandingPageMinimapButton:HookScript("OnEnter", showReportTooltip)
	GarrisonLandingPageMinimapButton:HookScript("OnLeave", hideReportTooltip)

	-- Update building, work order, follower, talent progress
--[[
	WARNING: LandingShipment is linked to OrderHall/CommandBar for tracking follower shipments
]]
	module:RegisterEvent("GARRISON_LANDINGPAGE_SHIPMENTS", trackAllShipments)

	-- Periodically update the GarrisonReport tooltip
	GarrisonLandingPageMinimapButton.UpdateTooltip = GarrisonLandingPageMinimapButton:GetScript("OnEnter")
--[[
	-- Track garrison cache across multiple characters
	module:On("GarrisonCacheLooted", trackGarrisonCache)

	-- Handle rush order
	module:On("RushOrderUsed", updateForRushOrder)]]

	-- Create character name for Garrison Report minimap button Alert
	GarrisonLandingPageMinimapButton.CharacterName = GarrisonLandingPageMinimapButton:CreateFontString(nil, "OVERLAY", "GameFontHighlightLeft")
	GarrisonLandingPageMinimapButton.CharacterName:SetPoint("TOPRIGHT", GarrisonLandingPageMinimapButton.AlertText, "BOTTOMRIGHT", 0, 12)
	GarrisonLandingPageMinimapButton.CharacterName:SetJustifyH("RIGHT")
	GarrisonLandingPageMinimapButton.CharacterName:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	GarrisonLandingPageMinimapButton.CharacterName:SetAlpha(0)
	GarrisonLandingPageMinimapButton.CharacterName:Hide()

	-- Animate charactername
	local fadeIn = GarrisonLandingPageMinimapButton.MinimapAlertAnim:CreateAnimation("Alpha")
	fadeIn:SetDuration(0.25)
	fadeIn:SetOrder(1)
	fadeIn:SetFromAlpha(0)
	fadeIn:SetToAlpha(1)
	fadeIn:SetChildKey("CharacterName")
	local fadeOut = GarrisonLandingPageMinimapButton.MinimapAlertAnim:CreateAnimation("Alpha")
	fadeOut:SetStartDelay(5)
	fadeOut:SetDuration(0.25)
	fadeOut:SetOrder(2)
	fadeOut:SetFromAlpha(1)
	fadeOut:SetToAlpha(0)
	fadeOut:SetChildKey("CharacterName")

	-- Clear and hide charactername
	GarrisonLandingPageMinimapButton.MinimapAlertAnim:HookScript("OnStop", hideCharacterName)
	GarrisonLandingPageMinimapButton.MinimapAlertAnim:HookScript("OnFinished", hideCharacterName)
	hooksecurefunc("GarrisonMinimapInvasion_ShowPulse", hideCharacterName)
	hooksecurefunc("GarrisonMinimapShipmentCreated_ShowPulse", hideCharacterName)

	-- Capture Garrison Resources changes
	--[[module:RegisterEvent("CURRENCY_DISPLAY_UPDATE", updateGarrisonCurrency)]]

	-- Start the minimap button alert timer for the next finished mission/building/talent/workorder/troop across all characters
	createAlertTimer()
end

function module.register(missionFrame)
	-- Create CharacterName fontstring
	--[[GarrisonLandingPageReport.CharacterName = GarrisonLandingPageReport:CreateFontString(nil, "BORDER", "GameFontNormal")
	GarrisonLandingPageReport.CharacterName:SetPoint("TOPLEFT", GarrisonLandingPage.HeaderBar, "BOTTOMLEFT", 26, 13)
	GarrisonLandingPageReport.CharacterName:Show()

	-- Reset seleccted player on LandingPage close
	GarrisonLandingPage:HookScript("OnHide", selectCurrentPlayer)

	-- Hide dropdown on LandingPage clsoe
	GarrisonLandingPageReport:HookScript("OnHide", function()
		if ( UIDropDownMenu_GetCurrentDropDown() == MissionMinderCharacterMenu ) then
			CloseDropDownMenus();
		end
	end);

	-- Track mission progress across multiple characters
	module:RegisterEvent("GARRISON_MISSION_STARTED", trackMission)
	module:On("MISSION_COMPLETED", untrackMission)]]

	-- Identify and cleanup orphaned missions (missions that are tracked, but were already completed outside of the addon)
	--Note: MissionList seems like the logical choice here but using MissionList actually causes us to erroneously
	--remove newly tracked misisons
	--hooksecurefunc(missionFrame.MissionTab.MissionList, "UpdateMissions", cleanupOrphanedMissions)
	hooksecurefunc(missionFrame.MissionTab.MissionList, "UpdateMissions", trackInProgressMissions)

	-- Track finished missions (mostly needed for when Mission Completion Orders are used)
	--[[module:RegisterEvent("GARRISON_MISSION_FINISHED", trackFinishedMissions)
	GarrisonMissionFrame:HookScript("OnShow", trackFinishedMissions)

	-- Handle mission button clicks across characters
	module.GarrisonLandingPageReportMission_OnClick = GarrisonLandingPageReportMission_OnClick
	GarrisonLandingPageReportMission_OnClick = handleMissionClicksForCharacters
	local missionButtons = GarrisonLandingPageReport.List.listScroll.buttons
	for i = 1, #missionButtons do
		missionButtons[i]:SetScript("OnClick", GarrisonLandingPageReportMission_OnClick)
	end

	GarrisonLandingPageReport.EmptyString = GarrisonLandingPageReport:CreateFontString(nil, "BORDER", "GameFontHighlightLarge")
	GarrisonLandingPageReport.EmptyString:SetPoint("BOTTOMLEFT", 60, 40)
	GarrisonLandingPageReport.EmptyString:SetPoint("TOPRIGHT", GarrisonLandingPageReport.List, "TOPLEFT", 0, 0)
	GarrisonLandingPageReport.EmptyString:SetTextColor(1, 0.82, 0)
	GarrisonLandingPageReport.EmptyString:Hide()

	-- Create disbaled fonts for mission tabs
	GarrisonLandingPageReport.InProgress:SetDisabledFontObject(GameFontDisableMed2)
	GarrisonLandingPageReport.Available:SetDisabledFontObject(GameFontDisableMed2)

	-- Update Shipments for the selected character
	module.GarrisonLandingPageReport_GetShipments = GarrisonLandingPageReport_GetShipments
	GarrisonLandingPageReport_GetShipments = loadShipments

	-- Update shipment tooltip for the selected character
	module.GarrisonLandingPageReportShipment_OnEnter = GarrisonLandingPageReportShipment_OnEnter
	GarrisonLandingPageReportShipment_OnEnter = showShipmentTooltipForSelectedCharacter
	local shipments = GarrisonLandingPageReport.Shipments
	for i = 1, #shipments do
		shipments[i]:SetScript("OnEnter", GarrisonLandingPageReportShipment_OnEnter);
	end

	-- Periodically update shipment tooltips
	for i = 1, #shipments do
		shipments[i].UpdateTooltip = GarrisonLandingPageReportShipment_OnEnter
	end

	-- Update missions for the selected character
	hooksecurefunc("GarrisonLandingPageReportList_UpdateItems", updateMissionsForSelectedCharacter)
	module.GarrisonLandingPageReportList_UpdateAvailable = GarrisonLandingPageReportList_UpdateAvailable
	GarrisonLandingPageReportList_UpdateAvailable = updateAvailableMissionsForSelectedCharacter
	module.GarrisonLandingPageReportList_Update = GarrisonLandingPageReportList_Update
	GarrisonLandingPageReportList_Update = updateInProgressMissionsForSelectedCharacter

	-- Update mission tooltip for the selected character
	module.GarrisonLandingPageReportMission_OnEnter = GarrisonLandingPageReportMission_OnEnter
	GarrisonLandingPageReportMission_OnEnter = updateMissionTooltipForSelectedCharacter
	for i=1, #missionButtons do
		missionButtons[i]:SetScript("OnEnter", GarrisonLandingPageReportMission_OnEnter)
	end

	-- Periodically upate mission tooltips
	for i=1, #missionButtons do
		missionButtons[i].UpdateTooltip = GarrisonLandingPageReportMission_OnEnter
	end

	-- Track Garrison Currency for the current character
	local events = Lib.events
	events:RegisterEvent("tracker", "CURRENCY_DISPLAY_UPDATE", updateGarrisonCurrency)

	--[===[@debug@
	for i=1, #missionButtons do
		local secondsRemaining = missionButtons[i]:CreateFontString(nil, "ARTWORK", "GameFontNormal")
		secondsRemaining:SetPoint("BOTTOMRIGHT", missionButtons[i], -4, 4)
		secondsRemaining:SetJustifyH("RIGHT")
		secondsRemaining:SetTextColor(1, 1, 1)
		missionButtons[i].Seconds = secondsRemaining
	end
	--@end-debug@]===]

	initialized = true
	tryVariablesLoaded()]]
end




module:RegisterEvent("GARRISON_BUILDING_PLACED", function() end)
module:RegisterEvent("GARRISON_BUILDING_REMOVED", function() end)
module:RegisterEvent("GARRISON_BUILDING_ACTIVATED", function() end)

module:RegisterEvent("SHIPMENT_CRAFTER_OPENED", function() end)
module:RegisterEvent("SHIPMENT_CRAFTER_INFO", function() end)
module:RegisterEvent("SHIPMENT_CRAFTER_CLOSED", function() end)

module:RegisterEvent("LOOT_CLOSED", function() end)


module:RegisterEvent("GARRISON_MISSION_NPC_OPENED", function() end)

--[[Lib.RegisterCallback("tracker.alert", "VARIABLES_LOADED", createAlertTimer)
Lib.RegisterCallback("tracker.currency", "VARIABLES_LOADED", updateGarrisonCurrency)
Lib:RegisterEvent("tracker", "ADDON_LOADED", initialize);]]

--[[local landingPageIsShown = false
Lib:RegisterEvent("tracker", "GARRISON_SHOW_LANDING_PAGE", function(...) landingPageIsShown = true; end)
Lib.RegisterCallback("tracker.hasGarrison", "VARIABLES_LOADED", function(...)
	local details = Lib.Globals.Characters[CURRENT_PLAYER]
	if ( not details.hasGarrison ) then
		details.hasGarrison = landingPageIsShown
	end
end)]]
